"""
This module provides tools for PyTorch optimizers.
"""

import typing
import torch


def _migrate_tensor(tensor: torch.Tensor, device: torch.device) -> None:
    """
    Migrates the tensor to the specified device.
    """
    tensor.data = tensor.data.to(device=device)
    if tensor.grad is not None:
        tensor.grad.data = tensor.grad.data.to(device=device)


def _migrate_param(param: object, device: torch.device) -> None:
    """
    Migrates the parameter to the specified device.
    """
    if isinstance(param, torch.Tensor):
        _migrate_tensor(param, device)
    elif isinstance(param, list):
        for subparam in param:
            _migrate_param(subparam, device)
    elif isinstance(param, dict):
        for subparam in param.values():
            _migrate_param(subparam, device)
    elif isinstance(param, int | float | complex):
        pass
    else:
        raise ValueError(f"Unexpected parameter type: {type(param)}")


def _migrate_optimizer(optimizer: torch.optim.Optimizer) -> None:
    """
    Migrates the optimizer to the device of the model parameters.
    """
    device: torch.device = optimizer.param_groups[0]["params"][0].device
    _migrate_param(optimizer.state, device)


def initialize_optimizer(  # pylint: disable=too-many-arguments
    params: typing.Iterable[torch.Tensor],
    *,
    use_lbfgs: bool,
    learning_rate: float,
    new_opt: bool = True,
    optimizer: torch.optim.Optimizer | None = None,
    state_dict: typing.Any = None,
) -> torch.optim.Optimizer:
    """
    Initialize an optimizer.

    Parameters
    ----------
    params : typing.Iterable[torch.Tensor]
        The parameters to be optimized.
    use_lbfgs : bool
        Whether to use L-BFGS as the optimizer.
    learning_rate : float
        The initial learning rate.
    new_opt : bool, default=True
        Whether to create a new optimizer or use the given optimizer, by default True.
    optimizer : torch.optim.Optimizer, optional
        The optimizer to be used if new_opt is False, by default None.
    state_dict : typing.Any, optional
        The state dictionary of the optimizer to be loaded, by default None.

    Returns
    -------
    torch.optim.Optimizer
        The optimizer.
    """
    if new_opt or optimizer is None:
        if use_lbfgs:
            optimizer = torch.optim.LBFGS(params, lr=learning_rate)
        else:
            optimizer = torch.optim.Adam(params, lr=learning_rate)
    if state_dict is not None:
        optimizer.load_state_dict(state_dict)
        _migrate_optimizer(optimizer)
    return optimizer


def scale_learning_rate(optimizer: torch.optim.Optimizer, scale: float) -> None:
    """
    Scales the learning rate of all parameter groups in the optimizer by a given factor.

    Parameters
    ----------
    optimizer : torch.optim.Optimizer
        The optimizer whose learning rate will be scaled.
    scale : float
        The factor by which the learning rate will be scaled.
    """
    for param in optimizer.param_groups:
        param["lr"] *= scale



"""
This file implements a perturbation estimator from haar.
"""

import logging
import typing
import dataclasses
import tyro
from .common import CommonConfig
from .subcommand_dict import subcommand_dict


@dataclasses.dataclass
class PerturbationConfig:
    """
    The perturbation estimator from haar.
    """

    common: typing.Annotated[CommonConfig, tyro.conf.OmitArgPrefixes]

    def main(self, *, model_param: typing.Any = None, network_param: typing.Any = None) -> None:
        """
        The main function of two-step optimization process based on imaginary time.
        """
        # pylint: disable=too-many-locals
        # pylint: disable=too-many-statements
        # pylint: disable=too-many-branches

        model, _, data = self.common.main(model_param=model_param, network_param=network_param)

        if "haar" not in data and "imag" in data:
            data["haar"] = data.pop("imag")
        configs, psi = data["haar"]["pool"]
        configs = configs.to(self.common.device)
        psi = psi.to(self.common.device)

        energy0_num = psi.conj() @ model.apply_within(configs, psi, configs)
        energy0_den = psi.conj() @ psi
        energy0 = (energy0_num / energy0_den).real.item()
        logging.info("Current energy is %.8f", energy0)
        logging.info("Reference energy is %.8f", model.ref_energy)

        number = configs.size(0)
        last_result_number = 0
        current_target_number = number
        logging.info("Starting finding relative configurations with %d.", number)
        while True:
            other_configs = model.find_relative(configs, psi, current_target_number, configs)
            current_result_number = other_configs.size(0)
            logging.info("Found %d relative configurations.", current_result_number)
            if current_result_number == last_result_number:
                logging.info("No new configurations found, stopping at %d.", current_result_number)
                break
            current_target_number = current_target_number * 2
            logging.info("Doubling target number to %d.", current_target_number)
            break

        hamiltonian_psi = model.apply_within(configs, psi, other_configs)
        energy2_num = (hamiltonian_psi.conj() @ hamiltonian_psi).real / (psi.conj() @ psi).real
        energy2_den = energy0 - model.diagonal_term(other_configs).real
        energy2 = (energy2_num / energy2_den).sum().item()
        logging.info("Correct energy is %.8f", energy2)
        logging.info("Error is reduced from %.8f to %.8f", energy0 - model.ref_energy, energy2 - model.ref_energy)


subcommand_dict["pert"] = PerturbationConfig



"""
This file implements the subspace chopping for the result of the imag script.
"""

import logging
import typing
import dataclasses
import tyro
import torch.utils.tensorboard
from .common import CommonConfig
from .subcommand_dict import subcommand_dict


@dataclasses.dataclass
class ChopImagConfig:
    """
    The subspace chopping for the result of the imag script.
    """

    common: typing.Annotated[CommonConfig, tyro.conf.OmitArgPrefixes]

    # The number of configurations to eliminate every iteration
    chop_size: typing.Annotated[int, tyro.conf.arg(aliases=["-c"])] = 10000
    # The estimated magnitude of the second order term
    second_order_magnitude: typing.Annotated[float, tyro.conf.arg(aliases=["-s"])] = 0.0

    def main(self) -> None:
        """
        The main function for the subspace chopping.
        """

        # pylint: disable=too-many-locals

        model, _, data = self.common.main()

        logging.info(
            "Arguments Summary: "
            "Chop Size: %d, "
            "Second Order Magnitude: %.10f",
            self.chop_size,
            self.second_order_magnitude,
        )

        configs, psi = data["imag"]["pool"]
        configs = configs.to(device=self.common.device)
        psi = psi.to(device=self.common.device)

        writer = torch.utils.tensorboard.SummaryWriter(log_dir=self.common.folder())  # type: ignore[no-untyped-call]

        original_configs = configs
        original_psi = psi
        ordered_configs: list[torch.Tensor] = []
        ordered_psi: list[torch.Tensor] = []
        mapping: dict[int, tuple[float, float]] = {}

        i = 0
        while True:
            num_configs = len(configs)
            logging.info("The number of configurations: %d", num_configs)
            writer.add_scalar("chop_imag/num_configs", num_configs, i)  # type: ignore[no-untyped-call]
            psi = psi / psi.norm()
            hamiltonian_psi = model.apply_within(configs, psi, configs)
            psi_hamiltonian_psi = (psi.conj() @ hamiltonian_psi).real
            energy = psi_hamiltonian_psi
            logging.info("The energy: %.10f, The energy error is %.10f", energy.item(), energy.item() - model.ref_energy)
            writer.add_scalar("chop_imag/energy", energy.item(), i)  # type: ignore[no-untyped-call]
            writer.add_scalar("chop_imag/error", energy.item() - model.ref_energy, i)  # type: ignore[no-untyped-call]
            writer.flush()  # type: ignore[no-untyped-call]
            mapping[num_configs] = (energy.item(), energy.item() - model.ref_energy)
            if self.second_order_magnitude >= 0:
                grad = hamiltonian_psi - psi_hamiltonian_psi * psi
                delta = -psi.conj() * grad
                real_delta = 2 * delta.real
                second_order = (psi.conj() * psi).real * self.second_order_magnitude
                rate = (real_delta + second_order).argsort()
            else:
                second_order = (psi.conj() * psi).real
                rate = second_order.argsort()
            unselected = rate[:self.chop_size]
            ordered_configs.append(configs[unselected])
            ordered_psi.append(psi[unselected])
            selected = rate[self.chop_size:]
            if len(selected) == 0:
                break
            configs = configs[selected]
            psi = psi[selected]
            i += 1

        data["chop_imag"] = {
            "ordered_configs": torch.cat(ordered_configs, dim=0),
            "ordered_psi": torch.cat(ordered_psi, dim=0),
            "original_configs": original_configs,
            "original_psi": original_psi,
            "mapping": mapping,
        }
        self.common.save(data, 0)


subcommand_dict["chop_imag"] = ChopImagConfig



"""
This file lists loss functions used in the direct optimization of wavefunction.
"""

import dataclasses
import inspect
import torch
from . import losses
from .subcommand_dict import subcommand_dict


@dataclasses.dataclass
class ListLossConfig:
    """
    The list of loss functions used in the direct optimization of wavefunction.
    """

    def main(self) -> None:
        """
        The main function for listing loss functions.
        """
        for name, member in inspect.getmembers(losses):
            if isinstance(member, torch.jit.ScriptFunction) and not name.startswith("_"):
                print(name, member.__doc__)


subcommand_dict["list_loss"] = ListLossConfig



"""
This file implements the reinforcement learning based subspace diagonalization algorithm.
"""

import logging
import typing
import dataclasses
import functools
import tyro
import scipy
import torch
from .common import CommonConfig
from .subcommand_dict import subcommand_dict
from .model_dict import ModelProto
from .optimizer import initialize_optimizer
from .bitspack import pack_int


def lanczos_energy(model: ModelProto, configs: torch.Tensor, step: int, threshold: float) -> tuple[float, torch.Tensor]:
    """
    Calculate the energy using the Lanczos method.
    """
    vector = torch.randn([configs.size(0)], dtype=torch.complex128, device=configs.device)

    v: list[torch.Tensor] = [vector / torch.linalg.norm(vector)]  # pylint: disable=not-callable
    alpha: list[torch.Tensor] = []
    beta: list[torch.Tensor] = []
    w: torch.Tensor
    w = model.apply_within(configs, v[-1], configs)
    alpha.append((w.conj() @ v[-1]).real)
    w = w - alpha[-1] * v[-1]
    i = 0
    while True:
        norm_w = torch.linalg.norm(w)  # pylint: disable=not-callable
        if norm_w < threshold:
            break
        beta.append(norm_w)
        v.append(w / beta[-1])
        w = model.apply_within(configs, v[-1], configs)
        alpha.append((w.conj() @ v[-1]).real)
        if i == step:
            break
        w = w - alpha[-1] * v[-1] - beta[-1] * v[-2]
        v[-2] = v[-2].cpu()  # v maybe very large, so we need to move it to CPU
        i += 1

    if len(beta) == 0:
        return alpha[0].item(), v[0]
    vals, vecs = scipy.linalg.eigh_tridiagonal(torch.stack(alpha, dim=0).cpu(), torch.stack(beta, dim=0).cpu(), lapack_driver="stebz", select="i", select_range=(0, 0))
    energy = torch.as_tensor(vals[0])
    result = functools.reduce(torch.add, (weight[0] * vector.to(device=configs.device) for weight, vector in zip(vecs, v)))
    return energy.item(), result


def config_contributions(model: ModelProto, base_configs: torch.Tensor, active_configs: torch.Tensor, state: torch.Tensor) -> torch.Tensor:
    """
    Calculate the config contribution for the given configurations pool of the ground state approximation.
    """
    base_state = state[:base_configs.size(0)]
    active_state = state[base_configs.size(0):]
    num = (active_state.conj() * model.apply_within(base_configs, base_state, active_configs)).real * 2
    den = (base_state.conj() @ base_state).real
    result = num / den
    return result


@dataclasses.dataclass
class RldiagConfig:
    """
    The reinforcement learning based subspace diagonalization algorithm.
    """

    # pylint: disable=too-many-instance-attributes

    common: typing.Annotated[CommonConfig, tyro.conf.OmitArgPrefixes]

    # The initial configuration for the first step, which is usually the Hatree-Fock state for quantum chemistry system
    initial_config: typing.Annotated[typing.Optional[str], tyro.conf.arg(aliases=["-i"])] = None
    # The maximum size of the configuration pool
    max_pool_size: typing.Annotated[int, tyro.conf.arg(aliases=["-n"])] = 32768
    # The learning rate for the local optimizer
    learning_rate: typing.Annotated[float, tyro.conf.arg(aliases=["-r"])] = 1e-3
    # The step of lanczos iteration for calculating the energy
    lanczos_step: typing.Annotated[int, tyro.conf.arg(aliases=["-k"])] = 64
    # The thereshold for the lanczos iteration
    lanczos_threshold: typing.Annotated[float, tyro.conf.arg(aliases=["-d"])] = 1e-8
    # The coefficient of configuration number for the sigma calculation
    alpha: typing.Annotated[float, tyro.conf.arg(aliases=["-a"])] = 0.0

    def main(self) -> None:
        """
        The main function for the reinforcement learning based subspace diagonalization algorithm.
        """

        # pylint: disable=too-many-statements
        # pylint: disable=too-many-locals
        # pylint: disable=too-many-branches

        model, network, data = self.common.main()

        logging.info(
            "Arguments Summary: "
            "Initial Configuration: %s, "
            "Learning Rate: %.10f, "
            "Lanczos step: %d, "
            "Lanczos threshold: %.10f, "
            "Alpha: %.10f",
            self.initial_config if self.initial_config is not None else "None",
            self.learning_rate,
            self.lanczos_step,
            self.lanczos_threshold,
            self.alpha,
        )

        optimizer = initialize_optimizer(
            network.parameters(),
            use_lbfgs=False,
            learning_rate=self.learning_rate,
            state_dict=data.get("optimizer"),
        )

        if self.initial_config is None:
            if "rldiag" not in data:  # pylint: disable=no-else-raise
                raise ValueError("The initial configuration is not set, please set it.")
            else:
                configs = data["rldiag"]["configs"].to(device=self.common.device)
        else:
            # The parameter initial_config accepts two formats:
            # 1. The 0/1 string, such as "11111100000000000000000000000011110000000000110000110000"
            # 2. The packed string, such as "63,0,0,192,3,48,12"
            if all(i in "01" for i in self.initial_config):
                # The 0/1 string
                configs = pack_int(
                    torch.tensor([[int(i) for i in self.initial_config]], dtype=torch.bool, device=self.common.device),
                    size=1,
                )
            else:
                # The packed string
                configs = torch.tensor([[int(i) for i in self.initial_config.split(",")]], dtype=torch.uint8, device=self.common.device)
            if "rldiag" not in data:
                data["rldiag"] = {"global": 0, "local": 0, "configs": configs}
            else:
                data["rldiag"]["configs"] = configs
                data["rldiag"]["local"] = 0

        writer = torch.utils.tensorboard.SummaryWriter(log_dir=self.common.folder())  # type: ignore[no-untyped-call]

        last_state = None
        while True:
            logging.info("Starting a new cycle")

            logging.info("Evaluating each configuration")
            score = network(configs)
            logging.info("All configurations are evaluated")

            logging.info("Applying the action")
            # |  old config pool  |
            # | pruned | remained | expanded |
            #          |   new config pool   |
            action = score.real >= -self.alpha
            _, topk = torch.topk(score.real, k=score.size(0) // 2, dim=0)
            action[topk] = True
            if score.size(0) > self.max_pool_size:
                _, topk = torch.topk(-score.real, k=score.size(0) - self.max_pool_size)
                action[topk] = False
            action[0] = True
            remained_configs = configs[action]
            pruned_configs = configs[torch.logical_not(action)]
            expanded_configs = model.single_relative(remained_configs)  # There are duplicated config here
            effective_expanded_configs, previous_to_effective = torch.unique(expanded_configs, dim=0, return_inverse=True)
            old_configs = torch.cat([remained_configs, pruned_configs])
            new_configs = torch.cat([remained_configs, effective_expanded_configs])
            configs = new_configs
            logging.info("Action has been applied")

            configs_size = configs.size(0)
            logging.info("Configuration pool size: %d", configs_size)
            writer.add_scalar("rldiag/configs/global", configs_size, data["rldiag"]["global"])  # type: ignore[no-untyped-call]
            writer.add_scalar("rldiag/configs/local", configs_size, data["rldiag"]["local"])  # type: ignore[no-untyped-call]

            if last_state is not None:
                old_state = last_state[torch.cat([action.nonzero()[:, 0], torch.logical_not(action).nonzero()[:, 0]])]
            else:
                _, old_state = lanczos_energy(model, old_configs, self.lanczos_step, self.lanczos_threshold)
            new_energy, new_state = lanczos_energy(model, configs, self.lanczos_step, self.lanczos_threshold)
            last_state = new_state
            energy = new_energy
            logging.info("Current energy is %.10f, Reference energy is %.10f, Energy error is %.10f", energy, model.ref_energy, energy - model.ref_energy)
            writer.add_scalar("rldiag/energy/state/global", energy, data["rldiag"]["global"])  # type: ignore[no-untyped-call]
            writer.add_scalar("rldiag/energy/state/local", energy, data["rldiag"]["local"])  # type: ignore[no-untyped-call]
            writer.add_scalar("rldiag/energy/error/global", energy - model.ref_energy, data["rldiag"]["global"])  # type: ignore[no-untyped-call]
            writer.add_scalar("rldiag/energy/error/local", energy - model.ref_energy, data["rldiag"]["local"])  # type: ignore[no-untyped-call]
            if "base" not in data["rldiag"]:
                # This is the first time to calculate the energy, which is usually the energy of the Hatree-Fock state for quantum chemistry system
                # This will not be flushed acrossing different cycle chains.
                data["rldiag"]["base"] = energy

            old_contribution = config_contributions(model, remained_configs, pruned_configs, old_state)
            effective_new_contribution = config_contributions(model, remained_configs, effective_expanded_configs, new_state)
            new_contribution = effective_new_contribution[previous_to_effective]
            contribution = torch.cat([old_contribution, new_contribution])
            configs_for_training = torch.cat([pruned_configs, remained_configs])
            with torch.enable_grad():  # type: ignore[no-untyped-call]
                score_for_training = network(configs_for_training)
                loss = torch.linalg.norm(score_for_training + contribution)  # pylint: disable=not-callable
                optimizer.zero_grad()
                loss.backward()
            optimizer.step()  # pylint: disable=no-value-for-parameter

            logging.info("Saving model checkpoint")
            data["rldiag"]["configs"] = configs
            data["rldiag"]["energy"] = energy
            data["rldiag"]["global"] += 1
            data["rldiag"]["local"] += 1
            data["network"] = network.state_dict()
            data["optimizer"] = optimizer.state_dict()
            self.common.save(data, data["rldiag"]["global"])
            logging.info("Checkpoint successfully saved")

            logging.info("Current cycle completed")


subcommand_dict["rldiag"] = RldiagConfig



"""
This file offers an interface for defining Hubbard models on a two-dimensional lattice.
"""

import typing
import logging
import dataclasses
import torch
import tyro
from .mlp import WaveFunctionElectronUpDown as MlpWaveFunction
from .attention import WaveFunctionElectronUpDown as AttentionWaveFunction
from .hamiltonian import Hamiltonian
from .model_dict import model_dict, ModelProto, NetworkProto, NetworkConfigProto


@dataclasses.dataclass
class ModelConfig:
    """
    The configuration for the Hubbard model.
    """

    # The width of the hubbard lattice
    m: typing.Annotated[int, tyro.conf.Positional]
    # The height of the hubbard lattice
    n: typing.Annotated[int, tyro.conf.Positional]

    # The coefficient of t
    t: typing.Annotated[float, tyro.conf.arg(aliases=["-t"])] = 1
    # The coefficient of U
    u: typing.Annotated[float, tyro.conf.arg(aliases=["-u"])] = 0

    # The electron number, left empty for half-filling
    electron_number: typing.Annotated[int | None, tyro.conf.arg(aliases=["-e"])] = None

    # The ref energy of the model
    ref_energy: typing.Annotated[float, tyro.conf.arg(aliases=["-r"])] = 0

    def __post_init__(self) -> None:
        if self.electron_number is None:
            self.electron_number = self.m * self.n

        if self.m <= 0 or self.n <= 0:
            raise ValueError("The dimensions of the Hubbard model must be positive integers.")

        if self.electron_number < 0 or self.electron_number > 2 * self.m * self.n:
            raise ValueError(f"The electron number {self.electron_number} is out of bounds for a {self.m}x{self.n} lattice. Each site can host up to two electrons (spin up and spin down).")


class Model(ModelProto[ModelConfig]):
    """
    This class handles the Hubbard model.
    """

    network_dict: dict[str, type[NetworkConfigProto["Model"]]] = {}

    config_t = ModelConfig

    @classmethod
    def default_group_name(cls, config: ModelConfig) -> str:
        return f"Hubbard_{config.m}x{config.n}_t{config.t}_u{config.u}_e{config.electron_number}"

    @classmethod
    def _prepare_hamiltonian(cls, args: ModelConfig) -> dict[tuple[tuple[int, int], ...], complex]:

        def _index(i: int, j: int, o: int) -> int:
            return (i + j * args.m) * 2 + o

        hamiltonian_dict: dict[tuple[tuple[int, int], ...], complex] = {}
        for i in range(args.m):
            for j in range(args.n):
                # On-site interaction
                hamiltonian_dict[(_index(i, j, 0), 1), (_index(i, j, 0), 0), (_index(i, j, 1), 1), (_index(i, j, 1), 0)] = args.u

                # Nearest neighbor hopping
                if i != 0:
                    hamiltonian_dict[(_index(i, j, 0), 1), (_index(i - 1, j, 0), 0)] = -args.t
                    hamiltonian_dict[(_index(i - 1, j, 0), 1), (_index(i, j, 0), 0)] = -args.t
                    hamiltonian_dict[(_index(i, j, 1), 1), (_index(i - 1, j, 1), 0)] = -args.t
                    hamiltonian_dict[(_index(i - 1, j, 1), 1), (_index(i, j, 1), 0)] = -args.t
                if j != 0:
                    hamiltonian_dict[(_index(i, j, 0), 1), (_index(i, j - 1, 0), 0)] = -args.t
                    hamiltonian_dict[(_index(i, j - 1, 0), 1), (_index(i, j, 0), 0)] = -args.t
                    hamiltonian_dict[(_index(i, j, 1), 1), (_index(i, j - 1, 1), 0)] = -args.t
                    hamiltonian_dict[(_index(i, j - 1, 1), 1), (_index(i, j, 1), 0)] = -args.t

        return hamiltonian_dict

    def __init__(self, args: ModelConfig):
        logging.info("Input arguments successfully parsed")

        assert args.electron_number is not None
        self.m: int = args.m
        self.n: int = args.n
        self.electron_number: int = args.electron_number
        logging.info("Constructing Hubbard model with dimensions: width = %d, height = %d", self.m, self.n)
        logging.info("The parameters of the model are: t = %.10f, U = %.10f, N = %d", args.t, args.u, args.electron_number)

        logging.info("Initializing Hamiltonian for the lattice")
        hamiltonian_dict: dict[tuple[tuple[int, int], ...], complex] = self._prepare_hamiltonian(args)
        logging.info("Hamiltonian initialization complete")

        self.ref_energy: float = args.ref_energy
        logging.info("The ref energy is set to %.10f", self.ref_energy)

        logging.info("Converting the Hamiltonian to internal Hamiltonian representation")
        self.hamiltonian: Hamiltonian = Hamiltonian(hamiltonian_dict, kind="fermi")
        logging.info("Internal Hamiltonian representation for model has been successfully created")

    def apply_within(self, configs_i: torch.Tensor, psi_i: torch.Tensor, configs_j: torch.Tensor) -> torch.Tensor:
        return self.hamiltonian.apply_within(configs_i, psi_i, configs_j)

    def find_relative(self, configs_i: torch.Tensor, psi_i: torch.Tensor, count_selected: int, configs_exclude: torch.Tensor | None = None) -> torch.Tensor:
        return self.hamiltonian.find_relative(configs_i, psi_i, count_selected, configs_exclude)

    def diagonal_term(self, configs: torch.Tensor) -> torch.Tensor:
        return self.hamiltonian.diagonal_term(configs)

    def single_relative(self, configs: torch.Tensor) -> torch.Tensor:
        return self.hamiltonian.single_relative(configs)

    def show_config(self, config: torch.Tensor) -> str:
        string = "".join(f"{i:08b}"[::-1] for i in config.cpu().numpy())
        return "[" + ".".join("".join(self._show_config_site(string[(i + j * self.m) * 2:(i + j * self.m) * 2 + 2]) for i in range(self.m)) for j in range(self.n)) + "]"

    def _show_config_site(self, string: str) -> str:
        match string:
            case "00":
                return " "
            case "10":
                return "↑"
            case "01":
                return "↓"
            case "11":
                return "↕"
            case _:
                raise ValueError(f"Invalid string: {string}")


model_dict["hubbard"] = Model


@dataclasses.dataclass
class MlpConfig:
    """
    The configuration of the MLP network.
    """

    # The hidden widths of the network
    hidden: typing.Annotated[tuple[int, ...], tyro.conf.arg(aliases=["-w"])] = (512,)

    def create(self, model: Model) -> NetworkProto:
        """
        Create a MLP network for the model.
        """
        logging.info("Hidden layer widths: %a", self.hidden)

        network = MlpWaveFunction(
            double_sites=model.m * model.n * 2,
            physical_dim=2,
            is_complex=True,
            spin_up=model.electron_number // 2,
            spin_down=model.electron_number - model.electron_number // 2,
            hidden_size=self.hidden,
            ordering=+1,
        )

        return network


Model.network_dict["mlp"] = MlpConfig


@dataclasses.dataclass
class AttentionConfig:
    """
    The configuration of the attention network.
    """

    # Embedding dimension
    embedding_dim: typing.Annotated[int, tyro.conf.arg(aliases=["-e"])] = 512
    # Heads number
    heads_num: typing.Annotated[int, tyro.conf.arg(aliases=["-m"])] = 8
    # Feedforward dimension
    feed_forward_dim: typing.Annotated[int, tyro.conf.arg(aliases=["-f"])] = 2048
    # Shared expert number
    shared_expert_num: typing.Annotated[int, tyro.conf.arg(aliases=["-s"])] = 1
    # Routed expert number
    routed_expert_num: typing.Annotated[int, tyro.conf.arg(aliases=["-r"])] = 0
    # Selected expert number
    selected_expert_num: typing.Annotated[int, tyro.conf.arg(aliases=["-c"])] = 0
    # Network depth
    depth: typing.Annotated[int, tyro.conf.arg(aliases=["-d"])] = 6

    def create(self, model: Model) -> NetworkProto:
        """
        Create an attention network for the model.
        """
        logging.info(
            "Attention network configuration: "
            "embedding dimension: %d, "
            "number of heads: %d, "
            "feed-forward dimension: %d, "
            "shared expert number: %d, "
            "routed expert number: %d, "
            "selected expert number: %d, "
            "depth: %d",
            self.embedding_dim,
            self.heads_num,
            self.feed_forward_dim,
            self.shared_expert_num,
            self.routed_expert_num,
            self.selected_expert_num,
            self.depth,
        )

        network = AttentionWaveFunction(
            double_sites=model.m * model.n * 2,
            physical_dim=2,
            is_complex=True,
            spin_up=model.electron_number // 2,
            spin_down=model.electron_number - model.electron_number // 2,
            embedding_dim=self.embedding_dim,
            heads_num=self.heads_num,
            feed_forward_dim=self.feed_forward_dim,
            shared_num=self.shared_expert_num,
            routed_num=self.routed_expert_num,
            selected_num=self.selected_expert_num,
            depth=self.depth,
            ordering=+1,
        )

        return network


Model.network_dict["attention"] = AttentionConfig



"""
This file implements an auto-regressive transformers network with the sampling method introduced in https://arxiv.org/pdf/2408.07625.
This network makes use of DeepSeekMoE architecture introduced in https://arxiv.org/pdf/2401.06066.
"""

import typing
import torch
from .bitspack import pack_int, unpack_int


class FeedForward(torch.nn.Module):
    """
    A feed-forward layer for transformer architectures.
    """

    def __init__(self, embedding_dim: int, hidden_dim: int) -> None:
        super().__init__()
        self.model: torch.nn.Module = torch.nn.Sequential(
            torch.nn.Linear(embedding_dim, hidden_dim),
            torch.nn.GELU(),
            torch.nn.Linear(hidden_dim, embedding_dim),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the feed-forward layer.
        """
        # x: batch * site * embedding
        x = self.model(x)
        # x: batch * site * embedding
        return x


class SelfAttention(torch.nn.Module):
    """
    Self-attention unit with support for key-value cache.
    """

    def __init__(self, embedding_dim: int, heads_num: int) -> None:
        super().__init__()

        self.heads_num: int = heads_num
        self.heads_dim: int = embedding_dim // heads_num
        assert self.heads_num * self.heads_dim == embedding_dim

        self.qkv: torch.nn.Module = torch.nn.Linear(embedding_dim, 3 * embedding_dim)
        self.out: torch.nn.Module = torch.nn.Linear(embedding_dim, embedding_dim)

    def forward(
        self,
        x: torch.Tensor,
        kv_cache: typing.Optional[tuple[torch.Tensor, torch.Tensor]],
        mask: typing.Optional[torch.Tensor],
    ) -> tuple[torch.Tensor, tuple[torch.Tensor, torch.Tensor]]:
        """
        Forward pass of the self-attention layer.
        """
        # x: batch * site * embedding
        batch_size, sites, embedding_dim = x.shape
        q, k, v = self.qkv(x).split(embedding_dim, dim=-1)
        # q, k, v: batch * site * embedding
        q = q.view([batch_size, sites, self.heads_num, self.heads_dim])
        k = k.view([batch_size, sites, self.heads_num, self.heads_dim])
        v = v.view([batch_size, sites, self.heads_num, self.heads_dim])
        # q, k, v: batch, site, heads_num, heads_dim
        q = q.transpose(1, 2)
        k = k.transpose(1, 2)
        v = v.transpose(1, 2)
        # q, k, v: batch, heads_num, site, heads_dim
        if kv_cache is not None:
            k_cache, v_cache = kv_cache
            k = torch.cat([k_cache, k], dim=2)
            v = torch.cat([v_cache, v], dim=2)
        # q: batch, heads_num, site, heads_dim
        # k, v: batch, heads_num, total_site, heads_dim
        if mask is None:
            total_sites = k.shape[-2]
            mask = torch.ones(sites, total_sites, dtype=torch.bool, device=x.device).tril(diagonal=total_sites - sites)
        # call scaled dot product attention
        attn = torch.nn.functional.scaled_dot_product_attention(q, k, v, attn_mask=mask)  # pylint: disable=not-callable
        # attn: batch, heads_num, site, heads_dim
        out = attn.transpose(1, 2).reshape([batch_size, sites, embedding_dim])
        # out: batch, site, embedding_dim
        return self.out(out), (k, v)


class DecoderUnit(torch.nn.Module):
    """
    A decoder unit within the transformer architecture, integrating both self-attention and feed-forward layers.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(  # pylint: disable=too-many-arguments
            self,
            *,
            embedding_dim: int,
            heads_num: int,
            feed_forward_dim: int,
            shared_num: int,
            routed_num: int,
            selected_num: int,
    ) -> None:
        super().__init__()
        self.shared_num = shared_num
        self.routed_num = routed_num
        self.selected_num = selected_num

        self.attention: torch.nn.Module = SelfAttention(embedding_dim, heads_num)
        self.norm1: torch.nn.Module = torch.nn.LayerNorm(embedding_dim)
        self.feed_forward_shared: torch.nn.ModuleList = torch.nn.ModuleList([FeedForward(embedding_dim, feed_forward_dim) for _ in range(shared_num)])
        self.feed_forward_routed: torch.nn.ModuleList = torch.nn.ModuleList([FeedForward(embedding_dim, feed_forward_dim) for _ in range(routed_num)])
        self.centroid: torch.nn.Parameter = torch.nn.Parameter(torch.randn(routed_num, embedding_dim))
        self.norm2: torch.nn.Module = torch.nn.LayerNorm(embedding_dim)

        self.bias: torch.Tensor
        self.register_buffer("bias", torch.zeros([self.routed_num]))
        self.accumulater: torch.Tensor
        self.register_buffer("accumulater", torch.zeros([self.routed_num]))
        self.count: torch.Tensor
        self.register_buffer("count", torch.zeros([]))

    def forward(
        self,
        x: torch.Tensor,
        kv_cache: typing.Optional[tuple[torch.Tensor, torch.Tensor]],
        mask: typing.Optional[torch.Tensor],
    ) -> tuple[torch.Tensor, tuple[torch.Tensor, torch.Tensor]]:
        """
        Forward pass of the decoder unit.
        """
        # x, y: batch * site * embedding
        y, result_cache = self.attention(x, kv_cache, mask)
        x = self.norm1(x + y)

        # The following segmental code implements the DeepSeekMoE architecture.
        y = x
        for i, expert in enumerate(self.feed_forward_shared):
            y = y + expert(x)
        # similarity: batch * site * routed
        similarity = torch.nn.functional.softmax(x @ self.centroid.t(), dim=-1)
        # top_k_indices: batch * site * selected
        _, top_k_indices = torch.topk(similarity + self.bias, self.selected_num, dim=-1)
        # gate_prime, gate: batch * site * routed
        gate_prime = torch.zeros_like(similarity).scatter_(-1, top_k_indices, similarity.gather(-1, top_k_indices))
        gate = gate_prime / gate_prime.sum(dim=-1).unsqueeze(-1)
        for i, expert in enumerate(self.feed_forward_routed):
            y = y + expert(x) * gate[:, :, i].unsqueeze(-1)
        x = self.norm2(y)

        if self.training:
            self.accumulater = self.accumulater + similarity.sum([0, 1])
            self.count = self.count + similarity.size(0) * similarity.size(1)

        return x, result_cache


class Transformers(torch.nn.Module):
    """
    A transformer model consisting of multiple decoder units.
    """

    def __init__(  # pylint: disable=too-many-arguments
            self,
            *,
            embedding_dim: int,
            heads_num: int,
            feed_forward_dim: int,
            shared_num: int,
            routed_num: int,
            selected_num: int,
            depth: int,
    ) -> None:
        super().__init__()
        self.layers: torch.nn.ModuleList = torch.nn.ModuleList(
            DecoderUnit(
                embedding_dim=embedding_dim,
                heads_num=heads_num,
                feed_forward_dim=feed_forward_dim,
                shared_num=shared_num,
                routed_num=routed_num,
                selected_num=selected_num,
            ) for _ in range(depth))

    def forward(
        self,
        x: torch.Tensor,
        kv_cache: typing.Optional[list[tuple[torch.Tensor, torch.Tensor]]],
        mask: typing.Optional[torch.Tensor],
    ) -> tuple[torch.Tensor, list[tuple[torch.Tensor, torch.Tensor]]]:
        """
        Forward pass of the transformer model.
        """
        # x: batch * site * embedding
        result_cache: list[tuple[torch.Tensor, torch.Tensor]] = []
        for i, layer in enumerate(self.layers):
            if kv_cache is None:
                x, cache = layer(x, None, mask)
            else:
                x, cache = layer(x, kv_cache[i], mask)
            result_cache.append(cache)
        return x, result_cache


class Tail(torch.nn.Module):
    """
    The Tail layer for the transformer model, responsible for mapping the final embeddings to the desired output dimensions.
    """

    def __init__(self, embedding_dim: int, hidden_dim: int, output_dim: int) -> None:
        super().__init__()
        self.model: torch.nn.Module = torch.nn.Sequential(
            torch.nn.Linear(embedding_dim, hidden_dim),
            torch.nn.GELU(),
            torch.nn.Linear(hidden_dim, output_dim),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the Tail layer.
        """
        # x: batch * site * embedding_dim
        x = self.model(x)
        # x: batch * site * output_dim
        return x


class Embedding(torch.nn.Module):
    """
    Embedding layer for transforming input data into a format suitable for the transformer model.
    """

    def __init__(self, sites: int, physical_dim: int, embedding_dim: int) -> None:
        super().__init__()
        self.parameter: torch.nn.Parameter = torch.nn.Parameter(torch.randn([sites, physical_dim, embedding_dim]))
        self.sites: int = sites
        self.physical_dim: int = physical_dim
        self.embedding_dim: int = embedding_dim

    def forward(self, x: torch.Tensor, base: int) -> torch.Tensor:
        """
        Forward pass of the Embedding layer.
        """
        # x: batch * sites
        batch, sites = x.shape

        x = x.unsqueeze(-1).unsqueeze(-1).expand(-1, -1, -1, self.embedding_dim)
        # x: batch * sites * config=1 * embedding

        parameter = self.parameter[base:base + sites].unsqueeze(0).expand(batch, -1, -1, -1)
        # parameter: batch * sites * config * embedding

        result = torch.gather(parameter, -2, x.to(dtype=torch.int64))
        # result: batch * site * 1 * embedding

        return result.squeeze(-2)


class WaveFunctionElectronUpDown(torch.nn.Module):
    """
    The wave function for the transformers network.
    This module maintains the conservation of particle number of spin-up and spin-down electrons.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(  # pylint: disable=too-many-arguments
            self,
            *,
            double_sites: int,  # Number of qubits, where each pair of qubits represents a site
            physical_dim: int,  # Dimension of the physical space, which is always 2
            is_complex: bool,  # Indicates whether the wave function is complex-valued, which is always true
            spin_up: int,  # Number of spin-up electrons
            spin_down: int,  # Number of spin-down electrons
            embedding_dim: int,  # Dimension of the embedding space used in the transformer layers
            heads_num: int,  # Number of attention heads in the transformer's self-attention mechanism
            feed_forward_dim: int,  # Dimension of the feed-forward network within the transformer layers
            shared_num: int,  # Number of the shared expert in the DeepSeekMoE architecture
            routed_num: int,  # Number of the routed expert in the DeepSeekMoE architecture
            selected_num: int,  # Number of the selected expert in the DeepSeekMoE architecture
            depth: int,  # Number of decoder layers in the transformer model
            ordering: int | list[int],  # Ordering of sites: +1 for normal order, -1 for reversed order, or a custom order list
    ) -> None:
        super().__init__()
        assert double_sites % 2 == 0
        self.double_sites: int = double_sites
        self.sites: int = double_sites // 2
        assert physical_dim == 2
        assert is_complex == True  # pylint: disable=singleton-comparison
        self.spin_up: int = spin_up
        self.spin_down: int = spin_down
        self.embedding_dim: int = embedding_dim
        self.heads_num: int = heads_num
        self.feed_forward_dim: int = feed_forward_dim
        self.shared_num: int = shared_num
        self.routed_num: int = routed_num
        self.selected_num: int = selected_num
        self.depth: int = depth

        # Embed configurations for each site, considering the four possible states of two qubits.
        self.embedding: torch.nn.Module = Embedding(self.sites, 4, self.embedding_dim)
        # Main body of the wave function computation.
        self.transformers: torch.nn.Module = Transformers(
            embedding_dim=self.embedding_dim,
            heads_num=self.heads_num,
            feed_forward_dim=self.feed_forward_dim,
            shared_num=self.shared_num,
            routed_num=self.routed_num,
            selected_num=self.selected_num,
            depth=self.depth,
        )
        # Tail layer mapping from embedding space to amplitude and phase space.
        # (amplitude and phase) * (4 possible states)
        self.tail: torch.nn.Module = Tail(self.embedding_dim, self.feed_forward_dim, 8)

        # Site Ordering Configuration
        # +1 for normal order, -1 for reversed order
        if isinstance(ordering, int) and ordering == +1:
            ordering = list(range(self.sites))
        if isinstance(ordering, int) and ordering == -1:
            ordering = list(reversed(range(self.sites)))
        self.ordering: torch.Tensor
        self.register_buffer("ordering", torch.tensor(ordering, dtype=torch.int64))
        self.ordering_reversed: torch.Tensor
        self.register_buffer("ordering_reversed", torch.scatter(torch.zeros(self.sites, dtype=torch.int64), 0, self.ordering, torch.arange(self.sites, dtype=torch.int64)))

        # Dummy Parameter for Device and Dtype Retrieval
        # This parameter is used to infer the device and dtype of the model.
        self.dummy_param = torch.nn.Parameter(torch.empty(0))

    @torch.jit.export
    def _mask(self, x: torch.Tensor) -> torch.Tensor:
        """
        Determined whether we could append spin up or spin down after uncompleted configurations.
        """
        # pylint: disable=too-many-locals

        # x: batch_size * current_site * 2
        # x represents the uncompleted configurations
        current_site = x.shape[1]
        # number: batch_size * 2
        # number denotes the total electron count for each uncompleted configuration
        number = x.sum(dim=1)

        # up/down_electron/hole: batch_size
        # These variables store the count of electrons and holes for each uncompleted configuration
        up_electron = number[:, 0]
        down_electron = number[:, 1]
        up_hole = current_site - up_electron
        down_hole = current_site - down_electron

        # add_up/down_electron/hole: batch_size
        # These variables determine whether it is possible to append an up/down electron/hole
        add_up_electron = up_electron < self.spin_up
        add_down_electron = down_electron < self.spin_down
        add_up_hole = up_hole < self.sites - self.spin_up
        add_down_hole = down_hole < self.sites - self.spin_down

        # add_up: batch_size * 2 * 1
        # add_down: batch_size * 1 * 2
        # These tensors represent the feasibility of adding up/down electrons/holes
        add_up = torch.stack([add_up_hole, add_up_electron], dim=-1).unsqueeze(-1)
        add_down = torch.stack([add_down_hole, add_down_electron], dim=-1).unsqueeze(-2)

        # add: batch_size * 2 * 2
        # add represents the logical AND of add_up and add_down, indicating the feasibility of appending specific electron/hole combinations
        # add[_, 0, 0] indicates the possibility of adding an up hole and a down hole
        # add[_, 0, 1] indicates the possibility of adding an up hole and a down electron
        # add[_, 1, 0] indicates the possibility of adding an up electron and a down hole
        # add[_, 1, 1] indicates the possibility of adding an up electron and a down electron
        add = torch.logical_and(add_up, add_down)

        return add

    @torch.jit.export
    def _normalize_amplitude(self, x: torch.Tensor) -> torch.Tensor:
        """
        Normalize the log amplitude of uncompleted configurations.
        """
        # x : ... * 2 * 2
        # param :  ...
        param = (2 * x).exp().sum(dim=[-2, -1]).log() / 2
        x = x - param.unsqueeze(-1).unsqueeze(-1)
        # 1 = param = sqrt(sum(x.exp()^2)) now
        return x

    @torch.jit.export
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Compute the wave function psi for the given configurations.
        """
        device: torch.device = self.dummy_param.device

        batch_size: int = x.shape[0]
        # x : batch_size * sites * 2
        x = unpack_int(x, size=1, last_dim=self.double_sites).view([batch_size, self.sites, 2])
        # Apply ordering
        x = torch.index_select(x, 1, self.ordering_reversed)

        # x4: batch_size * sites
        # Prepare x4 as the input for the neural network
        # The last configuration is excluded, and a zero vector is prepended at the beginning to represent the initial state.
        x4: torch.Tensor = x[:, :-1, 0] * 2 + x[:, :-1, 1]
        x4 = torch.cat([torch.zeros([batch_size, 1], device=device, dtype=torch.uint8), x4], dim=1)

        # emb: batch_size * sites * embedding
        # Embed the input tensor `x4` using the embedding layer.
        # In embedding layer, 0 for bos, 1 for site 0, ...
        emb: torch.Tensor = self.embedding(x4, 0)

        # post_transformers: batch_size * sites * embedding
        post_transformers, _ = self.transformers(emb, None, None)

        # tail: batch_size * sites * 8
        tail: torch.Tensor = self.tail(post_transformers)

        # amplitude/phase : batch_size * sites * 2 * 2
        amplitude: torch.Tensor = tail[:, :, :4].view(batch_size, self.sites, 2, 2)
        phase: torch.Tensor = tail[:, :, 4:].view(batch_size, self.sites, 2, 2)

        # Apply a filter mask to the amplitude to ensure the conservation of particle number.
        amplitude = amplitude + torch.stack([torch.where(self._mask(x[:, :i]), 0, -torch.inf) for i in range(self.sites)], dim=1)

        # Normalize the delta amplitude.
        amplitude = self._normalize_amplitude(amplitude)

        # batch/sites_indices: batch_size * sites
        batch_indices: torch.Tensor = torch.arange(batch_size).unsqueeze(1).expand(-1, self.sites)
        sites_indices: torch.Tensor = torch.arange(self.sites).unsqueeze(0).expand(batch_size, -1)

        # selected_amplitude/phase: batch_size * sites
        x_int32: torch.Tensor = x.to(dtype=torch.int32)
        selected_amplitude: torch.Tensor = amplitude[batch_indices, sites_indices, x_int32[:, :, 0], x_int32[:, :, 1]]
        selected_phase: torch.Tensor = phase[batch_indices, sites_indices, x_int32[:, :, 0], x_int32[:, :, 1]]

        return torch.view_as_complex(torch.stack([selected_amplitude.double().sum(dim=1), selected_phase.double().sum(dim=1)], dim=-1)).exp()

    @torch.jit.export
    def _blocked_forward_for_generate_unique(  # pylint: disable=too-many-arguments
        self,
        *,
        x: torch.Tensor,
        cache_input: typing.Optional[list[tuple[torch.Tensor, torch.Tensor]]],
        block_num: int,
        device: torch.device,
        i: int,
    ) -> tuple[torch.Tensor, list[tuple[torch.Tensor, torch.Tensor]]]:
        # pylint: disable=too-many-locals
        local_batch_size: int = x.size(0)
        local_batch_size_block = local_batch_size // block_num
        remainder = local_batch_size % block_num
        tail_list: list[torch.Tensor] = []
        cache_list: list[list[tuple[torch.Tensor, torch.Tensor]]] = []
        for j in range(block_num):
            if j < remainder:
                current_local_batch_size_block = local_batch_size_block + 1
            else:
                current_local_batch_size_block = local_batch_size_block
            start_index = j * local_batch_size_block + min(j, remainder)
            end_index = start_index + current_local_batch_size_block
            batch_indices_block = torch.arange(start_index, end_index, device=device, dtype=torch.int64)
            x_block: torch.Tensor = x[batch_indices_block]
            xi_block: torch.Tensor = x_block[:, -1:, :]
            xi4_block: torch.Tensor = xi_block[:, :, 0] * 2 + xi_block[:, :, 1]
            emb_block: torch.Tensor = self.embedding(xi4_block, i)
            if cache_input is None:
                post_transformers_block, cache_block = self.transformers(emb_block, None, None)
            else:
                cache_block_input = [(cache_layer[0][batch_indices_block], cache_layer[1][batch_indices_block]) for cache_layer in cache_input]
                post_transformers_block, cache_block = self.transformers(emb_block, cache_block_input, None)
            tail_block: torch.Tensor = self.tail(post_transformers_block)
            tail_list.append(tail_block)
            cache_list.append(cache_block)
        tail: torch.Tensor = torch.cat(tail_list)
        cache: list[tuple[torch.Tensor, torch.Tensor]] = [
            (torch.cat([cache_block[depth][0] for cache_block in cache_list]), torch.cat([cache_block[depth][1] for cache_block in cache_list])) for depth in range(self.depth)
        ]
        return tail, cache

    @torch.jit.export
    def generate_unique(self, batch_size: int, block_num: int = 1) -> tuple[torch.Tensor, torch.Tensor, None, None]:
        """
        Generate configurations uniquely.
        see https://arxiv.org/pdf/2408.07625.
        """
        # pylint: disable=too-many-locals
        # pylint: disable=invalid-name
        # pylint: disable=too-many-statements

        device: torch.device = self.dummy_param.device
        dtype: torch.dtype = self.dummy_param.dtype

        cache: typing.Optional[list[tuple[torch.Tensor, torch.Tensor]]] = None

        # x: local_batch_size * current_site * 2
        x: torch.Tensor = torch.zeros([1, 1, 2], device=device, dtype=torch.uint8)  # site=1, since the first is bos
        # (un)perturbed_log_probability: local_batch_size
        unperturbed_probability: torch.Tensor = torch.tensor([0], dtype=dtype, device=device)
        perturbed_probability: torch.Tensor = torch.tensor([0], dtype=dtype, device=device)
        for i in range(self.sites):
            local_batch_size: int = x.size(0)

            # # xi: batch * (sites=1) * 2
            # xi: torch.Tensor = x[:, -1:, :]
            # # xi4: batch * (sites=1)
            # xi4: torch.Tensor = xi[:, :, 0] * 2 + xi[:, :, 1]
            # # emb: batch * (sites=1) * embedding
            # emb: torch.Tensor = self.embedding(xi4, i)
            # # post_transformers: batch * (sites=1) * embedding
            # post_transformers, cache = self.transformers(emb, cache, None)
            # # tail: batch * (sites=1) * 8
            # tail: torch.Tensor = self.tail(post_transformers)
            tail, cache = self._blocked_forward_for_generate_unique(x=x, cache_input=cache, block_num=block_num, device=device, i=i)

            # The first 4 item are amplitude
            # delta_amplitude: batch * 2 * 2
            # delta_amplitude represents the amplitude changes for the configurations at the new site.
            delta_amplitude: torch.Tensor = tail[:, :, :4].view([local_batch_size, 2, 2])
            # Apply a filter mask to the amplitude to ensure the conservation of particle number.
            delta_amplitude = delta_amplitude + torch.where(self._mask(x[:, 1:]), 0, -torch.inf)

            # normalized_delta_amplitude: batch_size * 2 * 2
            # Normalize the delta amplitude.
            normalized_delta_amplitude: torch.Tensor = self._normalize_amplitude(delta_amplitude)

            # The delta unperturbed prob for all batch and 4 adds
            l: torch.Tensor = (2 * normalized_delta_amplitude).view([local_batch_size, 4])
            # and add to get the current unperturbed prob
            l = unperturbed_probability.view([local_batch_size, 1]) + l
            # Get perturbed prob by adding GUMBEL(0)
            L: torch.Tensor = l - (-torch.rand_like(l).log()).log()
            # Get max perturbed prob
            Z: torch.Tensor = L.max(dim=-1).values.view([local_batch_size, 1])
            # Evaluate the conditioned prob
            tildeL: torch.Tensor = -torch.log(torch.exp(-perturbed_probability.view([local_batch_size, 1])) - torch.exp(-Z) + torch.exp(-L))

            assert cache is not None

            # Calculate appended configurations for 4 adds
            # local_batch_size * current_site * 2 + local_batch_size * 1 * 2
            x0: torch.Tensor = torch.cat([x, torch.tensor([[0, 0]], device=device, dtype=torch.uint8).expand(local_batch_size, -1, -1)], dim=1)
            x1: torch.Tensor = torch.cat([x, torch.tensor([[0, 1]], device=device, dtype=torch.uint8).expand(local_batch_size, -1, -1)], dim=1)
            x2: torch.Tensor = torch.cat([x, torch.tensor([[1, 0]], device=device, dtype=torch.uint8).expand(local_batch_size, -1, -1)], dim=1)
            x3: torch.Tensor = torch.cat([x, torch.tensor([[1, 1]], device=device, dtype=torch.uint8).expand(local_batch_size, -1, -1)], dim=1)

            # Cat all configurations to get x : new_local_batch_size * (current_size+1) * 2
            # (un)perturbed prob : new_local_batch_size
            x = torch.cat([x0, x1, x2, x3])
            unperturbed_probability = l.permute(1, 0).reshape([4 * local_batch_size])
            perturbed_probability = tildeL.permute(1, 0).reshape([4 * local_batch_size])
            cache_indices = torch.arange(local_batch_size, device=device, dtype=torch.int64).unsqueeze(0).expand(4, -1).reshape([4 * local_batch_size])
            # kv cache: batch * heads_num * site * heads_dim, so just repeat first dimension

            # Filter results, only use largest batch_size ones
            selected = perturbed_probability.argsort(descending=True)[:batch_size]
            x = x[selected]
            unperturbed_probability = unperturbed_probability[selected]
            perturbed_probability = perturbed_probability[selected]
            cache_indices = cache_indices[selected]

            # If prob = 0, filter it forcely
            selected = perturbed_probability.isfinite()
            x = x[selected]
            unperturbed_probability = unperturbed_probability[selected]
            perturbed_probability = perturbed_probability[selected]
            cache_indices = cache_indices[selected]

            cache = [(cache_layer[0][cache_indices], cache_layer[1][cache_indices]) for cache_layer in cache]

        # Apply ordering
        x = torch.index_select(x[:, 1:], 1, self.ordering)
        # Flatten site part of x
        x = x.view([x.size(0), self.double_sites])
        # It should return configurations, amplitudes, probabilities and multiplicities.
        # But it is unique generator, so the last two fields are None
        x = pack_int(x, size=1)

        # Calculate the amplitude for the generated configurations in the batch.
        real_batch_size = len(x)
        real_batch_size_block = real_batch_size // block_num
        remainder = real_batch_size % block_num
        amplitude_list = []
        for j in range(block_num):
            if j < remainder:
                current_real_batch_size_block = real_batch_size_block + 1
            else:
                current_real_batch_size_block = real_batch_size_block
            start_index = j * real_batch_size_block + min(j, remainder)
            end_index = start_index + current_real_batch_size_block
            batch_indices_block = torch.arange(start_index, end_index, device=device, dtype=torch.int64)
            amplitude_block = self(x[batch_indices_block])
            amplitude_list.append(amplitude_block)
        amplitude: torch.Tensor = torch.cat(amplitude_list)

        return x, amplitude, None, None


class WaveFunctionNormal(torch.nn.Module):
    """
    The wave function for the transformers model.
    This module does not maintain any conservation.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(  # pylint: disable=too-many-arguments
            self,
            *,
            sites: int,  # Number of qubits
            physical_dim: int,  # Dimension of the physical space,
            is_complex: bool,  # Indicates whether the wave function is complex-valued, which is always true
            embedding_dim: int,  # Dimension of the embedding space used in the transformer layers
            heads_num: int,  # Number of attention heads in the transformer's self-attention mechanism
            feed_forward_dim: int,  # Dimension of the feed-forward network within the transformer layers
            shared_num: int,  # Number of the shared expert in the DeepSeekMoE architecture
            routed_num: int,  # Number of the routed expert in the DeepSeekMoE architecture
            selected_num: int,  # Number of the selected expert in the DeepSeekMoE architecture
            depth: int,  # Number of decoder layers in the transformer model
            ordering: int | list[int],  # Ordering of sites: +1 for normal order, -1 for reversed order, or a custom order list
    ) -> None:
        super().__init__()
        self.sites: int = sites
        self.physical_dim: int = physical_dim
        assert is_complex == True  # pylint: disable=singleton-comparison
        self.embedding_dim: int = embedding_dim
        self.heads_num: int = heads_num
        self.feed_forward_dim: int = feed_forward_dim
        self.shared_num: int = shared_num
        self.routed_num: int = routed_num
        self.selected_num: int = selected_num
        self.depth: int = depth

        # Embed configurations for each site, considering the physical_dim possible states of each qubit.
        self.embedding: torch.nn.Module = Embedding(self.sites, self.physical_dim, self.embedding_dim)
        # Main body of the wave function computation.
        self.transformers: torch.nn.Module = Transformers(
            embedding_dim=self.embedding_dim,
            heads_num=self.heads_num,
            feed_forward_dim=self.feed_forward_dim,
            shared_num=self.shared_num,
            routed_num=self.routed_num,
            selected_num=self.selected_num,
            depth=self.depth,
        )
        # Tail layer mapping from embedding space to amplitude and phase space.
        # (amplitude and phase) * (all possible states)
        self.tail: torch.nn.Module = Tail(self.embedding_dim, self.feed_forward_dim, 2 * self.physical_dim)

        # Site Ordering Configuration
        # +1 for normal order, -1 for reversed order
        if isinstance(ordering, int) and ordering == +1:
            ordering = list(range(self.sites))
        if isinstance(ordering, int) and ordering == -1:
            ordering = list(reversed(range(self.sites)))
        self.ordering: torch.Tensor
        self.register_buffer("ordering", torch.tensor(ordering, dtype=torch.int64))
        self.ordering_reversed: torch.Tensor
        self.register_buffer("ordering_reversed", torch.scatter(torch.zeros(self.sites, dtype=torch.int64), 0, self.ordering, torch.arange(self.sites, dtype=torch.int64)))

        # Dummy Parameter for Device and Dtype Retrieval
        # This parameter is used to infer the device and dtype of the model.
        self.dummy_param = torch.nn.Parameter(torch.empty(0))

    @torch.jit.export
    def _normalize_amplitude(self, x: torch.Tensor) -> torch.Tensor:
        """
        Normalize the log amplitude of uncompleted configurations.
        """
        # x : ... * physical_dim
        # param :  ...
        param = (2 * x).exp().sum(dim=[-1]).log() / 2
        x = x - param.unsqueeze(-1)
        # 1 = param = sqrt(sum(x.exp()^2)) now
        return x

    @torch.jit.export
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Compute the wave function psi for the given configurations.
        """
        device: torch.device = self.dummy_param.device

        batch_size: int = x.shape[0]
        # x : batch_size * sites
        x = unpack_int(x, size=self._bit_size(), last_dim=self.sites).view([batch_size, self.sites])
        # Apply ordering
        x = torch.index_select(x, 1, self.ordering_reversed)

        # x_for_emb: batch_size * sites
        # Prepare x for emb as the input for the neural network
        # The last configuration is excluded, and a zero vector is prepended at the beginning to represent the initial state.
        x_for_emb: torch.Tensor = x[:, :-1]
        x_for_emb = torch.cat([torch.zeros([batch_size, 1], device=device, dtype=torch.uint8), x_for_emb], dim=1)

        # emb: batch_size * sites * embedding
        # Embed the input tensor `x_for_emb` using the embedding layer.
        # In embedding layer, 0 for bos, 1 for site 0, ...
        emb: torch.Tensor = self.embedding(x_for_emb, 0)

        # post_transformers: batch_size * sites * embedding
        post_transformers, _ = self.transformers(emb, None, None)

        # tail: batch_size * sites * (2*physical_dim)
        tail: torch.Tensor = self.tail(post_transformers)

        # amplitude/phase : batch_size * sites * physical_dim
        amplitude: torch.Tensor = tail[:, :, :self.physical_dim]
        phase: torch.Tensor = tail[:, :, self.physical_dim:]

        # Normalize the delta amplitude.
        amplitude = self._normalize_amplitude(amplitude)

        # batch/sites_indices: batch_size * sites
        batch_indices: torch.Tensor = torch.arange(batch_size).unsqueeze(1).expand(-1, self.sites)
        sites_indices: torch.Tensor = torch.arange(self.sites).unsqueeze(0).expand(batch_size, -1)

        # selected_amplitude/phase: batch_size * sites
        x_int32: torch.Tensor = x.to(dtype=torch.int32)
        selected_amplitude: torch.Tensor = amplitude[batch_indices, sites_indices, x_int32]
        selected_phase: torch.Tensor = phase[batch_indices, sites_indices, x_int32]

        return torch.view_as_complex(torch.stack([selected_amplitude.double().sum(dim=1), selected_phase.double().sum(dim=1)], dim=-1)).exp()

    @torch.jit.export
    def _blocked_forward_for_generate_unique(  # pylint: disable=too-many-arguments
        self,
        *,
        x: torch.Tensor,
        cache_input: typing.Optional[list[tuple[torch.Tensor, torch.Tensor]]],
        block_num: int,
        device: torch.device,
        i: int,
    ) -> tuple[torch.Tensor, list[tuple[torch.Tensor, torch.Tensor]]]:
        # pylint: disable=too-many-locals
        local_batch_size: int = x.size(0)
        local_batch_size_block = local_batch_size // block_num
        remainder = local_batch_size % block_num
        tail_list: list[torch.Tensor] = []
        cache_list: list[list[tuple[torch.Tensor, torch.Tensor]]] = []
        for j in range(block_num):
            if j < remainder:
                current_local_batch_size_block = local_batch_size_block + 1
            else:
                current_local_batch_size_block = local_batch_size_block
            start_index = j * local_batch_size_block + min(j, remainder)
            end_index = start_index + current_local_batch_size_block
            batch_indices_block = torch.arange(start_index, end_index, device=device, dtype=torch.int64)
            x_block: torch.Tensor = x[batch_indices_block]
            xi_block: torch.Tensor = x_block[:, -1:]
            emb_block: torch.Tensor = self.embedding(xi_block, i)
            if cache_input is None:
                post_transformers_block, cache_block = self.transformers(emb_block, None, None)
            else:
                cache_block_input = [(cache_layer[0][batch_indices_block], cache_layer[1][batch_indices_block]) for cache_layer in cache_input]
                post_transformers_block, cache_block = self.transformers(emb_block, cache_block_input, None)
            tail_block: torch.Tensor = self.tail(post_transformers_block)
            tail_list.append(tail_block)
            cache_list.append(cache_block)
        tail: torch.Tensor = torch.cat(tail_list)
        cache: list[tuple[torch.Tensor, torch.Tensor]] = [
            (torch.cat([cache_block[depth][0] for cache_block in cache_list]), torch.cat([cache_block[depth][1] for cache_block in cache_list])) for depth in range(self.depth)
        ]
        return tail, cache

    @torch.jit.export
    def generate_unique(self, batch_size: int, block_num: int = 1) -> tuple[torch.Tensor, torch.Tensor, None, None]:
        """
        Generate configurations uniquely.
        see https://arxiv.org/pdf/2408.07625.
        """
        # pylint: disable=too-many-locals
        # pylint: disable=invalid-name
        # pylint: disable=too-many-statements

        device: torch.device = self.dummy_param.device
        dtype: torch.dtype = self.dummy_param.dtype

        cache: typing.Optional[list[tuple[torch.Tensor, torch.Tensor]]] = None

        # x: local_batch_size * current_site
        x: torch.Tensor = torch.zeros([1, 1], device=device, dtype=torch.uint8)  # site=1, since the first is bos
        # (un)perturbed_log_probability: local_batch_size
        unperturbed_probability: torch.Tensor = torch.tensor([0], dtype=dtype, device=device)
        perturbed_probability: torch.Tensor = torch.tensor([0], dtype=dtype, device=device)
        for i in range(self.sites):
            local_batch_size: int = x.size(0)

            # # xi: batch * (sites=1)
            # xi: torch.Tensor = x[:, -1:]
            # # emb: batch * (sites=1) * embedding
            # emb: torch.Tensor = self.embedding(xi, i)
            # # post_transformers: batch * (sites=1) * embedding
            # post_transformers, cache = self.transformers(emb, cache, None)
            # # tail: batch * (sites=1) * (2*physical_dim)
            # tail: torch.Tensor = self.tail(post_transformers)
            tail, cache = self._blocked_forward_for_generate_unique(x=x, cache_input=cache, block_num=block_num, device=device, i=i)

            # The first physical_dim item are amplitude
            # delta_amplitude: batch * physical_dim
            # delta_amplitude represents the amplitude changes for the configurations at the new site.
            delta_amplitude: torch.Tensor = tail[:, :, :self.physical_dim].view([local_batch_size, self.physical_dim])

            # normalized_delta_amplitude: batch_size * physical_dim
            # Normalize the delta amplitude.
            normalized_delta_amplitude: torch.Tensor = self._normalize_amplitude(delta_amplitude)

            # The delta unperturbed prob for all batch and all new possible states
            l: torch.Tensor = (2 * normalized_delta_amplitude).view([local_batch_size, self.physical_dim])
            # and add to get the current unperturbed prob
            l = unperturbed_probability.view([local_batch_size, 1]) + l
            # Get perturbed prob by adding GUMBEL(0)
            L: torch.Tensor = l - (-torch.rand_like(l).log()).log()
            # Get max perturbed prob
            Z: torch.Tensor = L.max(dim=-1).values.view([local_batch_size, 1])
            # Evaluate the conditioned prob
            tildeL: torch.Tensor = -torch.log(torch.exp(-perturbed_probability.view([local_batch_size, 1])) - torch.exp(-Z) + torch.exp(-L))

            assert cache is not None

            # Calculate appended configurations for all new possible states
            # local_batch_size * current_site + local_batch_size * 1
            xs: list[torch.Tensor] = [torch.cat([x, torch.tensor([j], device=device, dtype=torch.uint8).expand(local_batch_size, -1)], dim=1) for j in range(self.physical_dim)]

            # Cat all configurations to get x : new_local_batch_size * (current_size+1)
            # (un)perturbed prob : new_local_batch_size
            x = torch.cat(xs)
            unperturbed_probability = l.permute(1, 0).reshape([self.physical_dim * local_batch_size])
            perturbed_probability = tildeL.permute(1, 0).reshape([self.physical_dim * local_batch_size])
            cache_indices = torch.arange(local_batch_size, device=device, dtype=torch.int64).unsqueeze(0).expand(4, -1).reshape([4 * local_batch_size])
            # kv cache: batch * heads_num * site * heads_dim, so just repeat first dimension

            # Filter results, only use largest batch_size ones
            selected = perturbed_probability.argsort(descending=True)[:batch_size]
            x = x[selected]
            unperturbed_probability = unperturbed_probability[selected]
            perturbed_probability = perturbed_probability[selected]
            cache_indices = cache_indices[selected]

            # If prob = 0, filter it forcely
            selected = perturbed_probability.isfinite()
            x = x[selected]
            unperturbed_probability = unperturbed_probability[selected]
            perturbed_probability = perturbed_probability[selected]
            cache_indices = cache_indices[selected]

            cache = [(cache_layer[0][cache_indices], cache_layer[1][cache_indices]) for cache_layer in cache]

        # Apply ordering
        x = torch.index_select(x[:, 1:], 1, self.ordering)
        # Flatten site part of x
        x = x.view([x.size(0), self.sites])
        # It should return configurations, amplitudes, probabilities and multiplicities.
        # But it is unique generator, so the last two fields are None
        x = pack_int(x, size=self._bit_size())

        # Calculate the amplitude for the generated configurations in the batch.
        real_batch_size = len(x)
        real_batch_size_block = real_batch_size // block_num
        remainder = real_batch_size % block_num
        amplitude_list = []
        for j in range(block_num):
            if j < remainder:
                current_real_batch_size_block = real_batch_size_block + 1
            else:
                current_real_batch_size_block = real_batch_size_block
            start_index = j * real_batch_size_block + min(j, remainder)
            end_index = start_index + current_real_batch_size_block
            batch_indices_block = torch.arange(start_index, end_index, device=device, dtype=torch.int64)
            amplitude_block = self(x[batch_indices_block])
            amplitude_list.append(amplitude_block)
        amplitude: torch.Tensor = torch.cat(amplitude_list)

        return x, amplitude, None, None

    def _bit_size(self) -> int:
        if self.physical_dim <= 1 << 1:
            return 1
        if self.physical_dim <= 1 << 2:
            return 2
        if self.physical_dim <= 1 << 4:
            return 4
        if self.physical_dim <= 1 << 8:
            return 8
        raise ValueError("physical_dim should be less than or equal to 256")
